"""
Creates a mesh for a square with a round hole.
"""
import pygmsh
from helpers import compute_volume


def test():
    geom = pygmsh.built_in.Geometry()

    circle = geom.add_circle(
        x0=[0.5, 0.5, 0.0], radius=0.25, lcar=0.1, num_sections=4, make_surface=False
    )

    geom.add_rectangle(0.0, 1.0, 0.0, 1.0, 0.0, lcar=0.1, holes=[circle.line_loop])

    ref = 0.8086582838174551
    mesh = pygmsh.generate_mesh(geom, geo_filename="h.geo")
    assert abs(compute_volume(mesh) - ref) < 1.0e-2 * ref
    return mesh


if __name__ == "__main__":
    import meshio

    meshio.write("rectangle_with_hole.vtu", test())



import pygmsh
from helpers import compute_volume


def test():
    geom = pygmsh.built_in.Geometry()

    poly = geom.add_polygon(
        [
            [0.0, 0.0, 0.0],
            [2.0, 0.0, 0.0],
            [3.0, 1.0, 0.0],
            [1.0, 2.0, 0.0],
            [0.0, 1.0, 0.0],
        ],
        lcar=0.1,
    )

    field0 = geom.add_boundary_layer(
        edges_list=[poly.line_loop.lines[0]],
        hfar=0.1,
        hwall_n=0.01,
        ratio=1.1,
        thickness=0.2,
        anisomax=100.0,
    )

    field1 = geom.add_boundary_layer(
        nodes_list=[poly.line_loop.lines[1].points[1]],
        hfar=0.1,
        hwall_n=0.01,
        ratio=1.1,
        thickness=0.2,
        anisomax=100.0,
    )

    geom.add_background_field([field0, field1])

    ref = 4.0
    mesh = pygmsh.generate_mesh(geom)
    assert abs(compute_volume(mesh) - ref) < 1.0e-2 * ref
    return mesh


if __name__ == "__main__":
    import meshio

    meshio.write("boundary_layers.vtu", test())



class SurfaceBase:
    """
    Increments the Line ID every time a new object is created that inherits
    from LineBase.

    Parameters
    ----------
    id0 : str
        If no unique ID is given, the object global is incremented.
    """

    _ID = 0
    num_edges = 0
    dimension = 2

    def __init__(self, id0=None, num_edges=0):
        if id0:
            assert isinstance(id0, str)
            self.id = id0
        else:
            self.id = f"s{SurfaceBase._ID}"
            SurfaceBase._ID += 1
        self.num_edges = num_edges
        return



from ..__about__ import __version__
from ..built_in import geometry as bl
from .ball import Ball
from .box import Box
from .cone import Cone
from .cylinder import Cylinder
from .disk import Disk
from .ellipsoid import Ellipsoid
from .rectangle import Rectangle
from .surface_base import SurfaceBase
from .torus import Torus
from .volume_base import VolumeBase
from .wedge import Wedge


class Geometry(bl.Geometry):
    def __init__(self, characteristic_length_min=None, characteristic_length_max=None):
        super().__init__()
        self._BOOLEAN_ID = 0
        self._EXTRUDE_ID = 0
        self._GMSH_CODE = [
            f"// This code was created by pygmsh v{__version__}.",
            'SetFactory("OpenCASCADE");',
        ]

        if characteristic_length_min is not None:
            self._GMSH_CODE.append(
                f"Mesh.CharacteristicLengthMin = {characteristic_length_min};"
            )

        if characteristic_length_max is not None:
            self._GMSH_CODE.append(
                f"Mesh.CharacteristicLengthMax = {characteristic_length_max};"
            )
        return

    def get_code(self):
        """Returns properly formatted Gmsh code.
        """
        return "\n".join(self._GMSH_CODE)

    def add_rectangle(self, *args, **kwargs):
        p = Rectangle(*args, **kwargs)
        self._GMSH_CODE.append(p.code)
        return p

    def add_disk(self, *args, **kwargs):
        p = Disk(*args, **kwargs)
        self._GMSH_CODE.append(p.code)
        return p

    def add_ball(self, *args, **kwargs):
        p = Ball(*args, **kwargs)
        self._GMSH_CODE.append(p.code)
        return p

    def add_box(self, *args, **kwargs):
        p = Box(*args, **kwargs)
        self._GMSH_CODE.append(p.code)
        return p

    def add_cone(self, *args, **kwargs):
        p = Cone(*args, **kwargs)
        self._GMSH_CODE.append(p.code)
        return p

    def add_cylinder(self, *args, **kwargs):
        p = Cylinder(*args, **kwargs)
        self._GMSH_CODE.append(p.code)
        return p

    def add_ellipsoid(self, *args, **kwargs):
        p = Ellipsoid(*args, **kwargs)
        self._GMSH_CODE.append(p.code)
        return p

    def add_torus(self, *args, **kwargs):
        p = Torus(*args, **kwargs)
        self._GMSH_CODE.append(p.code)
        return p

    def add_wedge(self, *args, **kwargs):
        p = Wedge(*args, **kwargs)
        self._GMSH_CODE.append(p.code)
        return p

    def _boolean_operation(
        self,
        operation,
        input_entities,
        tool_entities,
        delete_first=True,
        delete_other=True,
    ):
        """Boolean operations, see
        https://gmsh.info/doc/texinfo/gmsh.html#Boolean-operations input_entity
        and tool_entity are called object and tool in gmsh documentation.
        """
        self._BOOLEAN_ID += 1

        # assert that all entities are of the same dimensionality
        dim = None
        legal_dim_types = {1: "Line", 2: "Surface", 3: "Volume"}
        for ldt in legal_dim_types:
            if input_entities[0].dimension == ldt:
                dim = ldt
                break
        assert dim is not None, "Illegal input type '{}' for Boolean operation.".format(
            type(input_entities[0])
        )
        for e in input_entities[1:] + tool_entities:
            assert (
                e.dimension == dim
            ), "Incompatible input type '{}' for Boolean operation.".format(type(e))

        name = f"bo{self._BOOLEAN_ID}"

        input_delete = "Delete;" if delete_first else ""

        tool_delete = "Delete;" if delete_other else ""

        legal_dim_type = legal_dim_types[dim]

        if input_entities:
            formatted_input_entities = (
                ";".join([f"{legal_dim_type}{{{e.id}}}" for e in input_entities]) + ";"
            )
        else:
            formatted_input_entities = ""

        if tool_entities:
            formatted_tool_entities = (
                ";".join([f"{legal_dim_type}{{{e.id}}}" for e in tool_entities]) + ";"
            )
        else:
            formatted_tool_entities = ""

        self._GMSH_CODE.append(
            # I wonder what this line does in Lisp. ;)
            # '{}[] = {}{{{} {{{}}}; {}}} {{{} {{{}}}; {}}};'
            # .format(
            #    name,
            #    operation,
            #    legal_dim_types[dim],
            #    ';'.join(e.id for e in input_entities),
            #    'Delete;' if delete_first else '',
            #    legal_dim_types[dim],
            #    ';'.join(e.id for e in tool_entities),
            #    'Delete;' if delete_other else ''
            #    ))
            "%(name)s[] = %(op)s{ %(ientities)s %(idelete)s } { %(tentities)s %(tdelete)s};"
            % {
                "name": name,
                "op": operation,
                "ientities": formatted_input_entities,
                "idelete": input_delete,
                "tentities": formatted_tool_entities,
                "tdelete": tool_delete,
            }
        )
        mapping = {"Line": None, "Surface": SurfaceBase, "Volume": VolumeBase}
        return mapping[legal_dim_types[dim]](id0=name, is_list=True)

    def boolean_intersection(self, entities, delete_first=True, delete_other=True):
        """Boolean intersection, see
        https://gmsh.info/doc/texinfo/gmsh.html#Boolean-operations input_entity
        and tool_entity are called object and tool in gmsh documentation.
        """
        assert len(entities) > 1
        return self._boolean_operation(
            "BooleanIntersection",
            [entities[0]],
            entities[1:],
            delete_first=delete_first,
            delete_other=delete_other,
        )

    def boolean_union(self, entities, delete_first=True, delete_other=True):
        """Boolean union, see
        https://gmsh.info/doc/texinfo/gmsh.html#Boolean-operations input_entity
        and tool_entity are called object and tool in gmsh documentation.
        """
        out = self._boolean_operation(
            "BooleanUnion",
            [entities[0]],
            entities[1:],
            delete_first=delete_first,
            delete_other=delete_other,
        )
        # Cannot add Compound Surface yet; see
        # <https://gitlab.onelab.info/gmsh/gmsh/issues/525>.
        # if compound:
        #     self._GMSH_CODE.append("Compound Surface {{{}}};".format(out.id))
        return out

    def boolean_difference(self, *args, **kwargs):
        """Boolean difference, see
        https://gmsh.info/doc/texinfo/gmsh.html#Boolean-operations input_entity
        and tool_entity are called object and tool in gmsh documentation.
        """
        return self._boolean_operation("BooleanDifference", *args, **kwargs)

    def boolean_fragments(self, *args, **kwargs):
        """Boolean fragments, see
        https://gmsh.info/doc/texinfo/gmsh.html#Boolean-operations input_entity
        and tool_entity are called object and tool in gmsh documentation.
        """
        return self._boolean_operation("BooleanFragments", *args, **kwargs)



import math

import numpy


def prune_nodes(points, cells):
    # Only points/cells that actually used
    uvertices, uidx = numpy.unique(cells, return_inverse=True)
    cells = uidx.reshape(cells.shape)
    points = points[uvertices]
    return points, cells


def get_triangle_volumes(pts, cells):
    # Works in any dimension; taken from voropy
    local_idx = numpy.array([[1, 2], [2, 0], [0, 1]]).T
    idx_hierarchy = cells.T[local_idx]

    half_edge_coords = pts[idx_hierarchy[1]] - pts[idx_hierarchy[0]]
    ei_dot_ej = numpy.einsum(
        "ijk, ijk->ij", half_edge_coords[[1, 2, 0]], half_edge_coords[[2, 0, 1]]
    )

    vols = 0.5 * numpy.sqrt(
        +ei_dot_ej[2] * ei_dot_ej[0]
        + ei_dot_ej[0] * ei_dot_ej[1]
        + ei_dot_ej[1] * ei_dot_ej[2]
    )
    return vols


def get_simplex_volumes(pts, cells):
    """Signed volume of a simplex in nD. Note that signing only makes sense for
    n-simplices in R^n.
    """
    n = pts.shape[1]
    assert cells.shape[1] == n + 1

    p = pts[cells]
    p = numpy.concatenate([p, numpy.ones(list(p.shape[:2]) + [1])], axis=-1)
    return numpy.abs(numpy.linalg.det(p) / math.factorial(n))


def compute_volume(mesh):
    if "tetra" in mesh.cells_dict:
        vol = math.fsum(
            get_simplex_volumes(*prune_nodes(mesh.points, mesh.cells_dict["tetra"]))
        )
    elif "triangle" in mesh.cells_dict or "quad" in mesh.cells_dict:
        vol = 0.0
        if "triangle" in mesh.cells_dict:
            # triangles
            vol += math.fsum(
                get_triangle_volumes(
                    *prune_nodes(mesh.points, mesh.cells_dict["triangle"])
                )
            )
        if "quad" in mesh.cells_dict:
            # quad: treat as two triangles
            quads = mesh.cells_dict["quad"].T
            split_cells = numpy.column_stack(
                [[quads[0], quads[1], quads[2]], [quads[0], quads[2], quads[3]]]
            ).T
            vol += math.fsum(
                get_triangle_volumes(*prune_nodes(mesh.points, split_cells))
            )
    else:
        assert "line" in mesh.cells_dict
        segs = numpy.diff(mesh.points[mesh.cells_dict["line"]], axis=1).squeeze()
        vol = numpy.sum(numpy.sqrt(numpy.einsum("...j, ...j", segs, segs)))

    return vol


def plot(filename, points, cells):
    import matplotlib.pyplot as plt

    pts = points[:, :2]
    for e in cells["triangle"]:
        for idx in [[0, 1], [1, 2], [2, 0]]:
            X = pts[e[idx]]
            plt.plot(X[:, 0], X[:, 1], "-k")
    plt.gca().set_aspect("equal", "datalim")
    plt.axis("off")

    # plt.show()
    plt.savefig(filename, transparent=True)
    return



import copy


class LineBase:
    """
    Increments the Line ID every time a new object
    is created that inherits from LineBase.

    Parameters
    ----------
    id0 : str
        If no unique ID is given, the object global is incremented.
    """

    _ID = 0
    dimension = 1

    def __init__(self, id0=None):
        if id0:
            self.id = id0
        else:
            self.id = f"l{LineBase._ID}"
            LineBase._ID += 1
        return

    def __neg__(self):
        neg_self = copy.deepcopy(self)
        neg_self.id = "-" + neg_self.id
        return neg_self



import pygmsh
from helpers import compute_volume


def test():
    geom = pygmsh.built_in.Geometry()

    geom.add_circle(
        [0.0, 0.0, 0.0],
        1.0,
        lcar=0.1,
        num_sections=4,
        # If compound==False, the section borders have to be points of the
        # discretization. If using a compound circle, they don't; gmsh can
        # choose by itself where to point the circle points.
        compound=True,
    )

    ref = 3.1363871677682247
    mesh = pygmsh.generate_mesh(geom, prune_z_0=True)
    assert abs(compute_volume(mesh) - ref) < 1.0e-2 * ref
    return mesh


if __name__ == "__main__":
    import meshio

    meshio.write("circle.vtk", test())



import os
import subprocess
import tempfile

import numpy

import meshio


def rotation_matrix(u, theta):
    """Return matrix that implements the rotation around the vector :math:`u`
    by the angle :math:`\\theta`, cf.
    https://en.wikipedia.org/wiki/Rotation_matrix#Rotation_matrix_from_axis_and_angle.

    :param u: rotation vector
    :param theta: rotation angle
    """
    assert numpy.isclose(numpy.inner(u, u), 1.0), "the rotation axis must be unitary"

    # Cross-product matrix.
    cpm = numpy.array([[0.0, -u[2], u[1]], [u[2], 0.0, -u[0]], [-u[1], u[0], 0.0]])
    c = numpy.cos(theta)
    s = numpy.sin(theta)
    R = numpy.eye(3) * c + s * cpm + (1.0 - c) * numpy.outer(u, u)
    return R


def _is_string(obj):
    try:
        # Python 2
        return isinstance(obj, basestring)
    except NameError:
        # Python 3
        return isinstance(obj, str)


def _get_gmsh_exe():
    macos_gmsh_location = "/Applications/Gmsh.app/Contents/MacOS/gmsh"
    return macos_gmsh_location if os.path.isfile(macos_gmsh_location) else "gmsh"


def get_gmsh_major_version(gmsh_exe=_get_gmsh_exe()):
    out = (
        subprocess.check_output([gmsh_exe, "--version"], stderr=subprocess.STDOUT)
        .strip()
        .decode("utf8")
    )
    ex = out.split(".")
    return int(ex[0])


def generate_mesh(  # noqa: C901
    geo_object,
    verbose=True,
    dim=3,
    prune_vertices=True,
    prune_z_0=False,
    remove_faces=False,
    gmsh_path=None,
    extra_gmsh_arguments=None,
    # for debugging purposes:
    geo_filename=None,
    msh_filename=None,
    mesh_file_type="msh",
):
    """Return a meshio.Mesh, storing the mesh points, cells, and data,
    generated by Gmsh from the `geo_object`, written to a temporary file,
    and reread by `meshio`.

    Gmsh's native "msh" format is ill-suited to fast I/O.  This can
    greatly reduce the performance of pygmsh.  As alternatives, try
    `mesh_file_type=`:

    - "vtk"`, though Gmsh doesn't write the physical tags
    to VTK <https://gitlab.onelab.info/gmsh/gmsh/issues/389> or

    - `"mesh"`, though this only supports a few basic elements - "line",
    "triangle", "quad", "tetra", "hexahedron" - and doesn't preserve
    the `$PhysicalNames`, just the `int` tags.

    """
    if extra_gmsh_arguments is None:
        extra_gmsh_arguments = []

    # For format "mesh", ask Gmsh to save the physical tags
    # http://gmsh.info/doc/texinfo/gmsh.html#index-Mesh_002eSaveElementTagType
    if mesh_file_type == "mesh":
        extra_gmsh_arguments += ["-string", "Mesh.SaveElementTagType=2;"]

    preserve_geo = geo_filename is not None
    if geo_filename is None:
        with tempfile.NamedTemporaryFile(suffix=".geo") as f:
            geo_filename = f.name

    with open(geo_filename, "w") as f:
        f.write(geo_object.get_code())

    # As of Gmsh 4.1.3, the mesh format options are
    # ```
    # auto, msh1, msh2, msh3, msh4, msh, unv, vtk, wrl, mail, stl, p3d, mesh, bdf, cgns,
    # med, diff, ir3, inp, ply2, celum, su2, x3d, dat, neu, m, key
    # ```
    # Pick the correct filename suffix.
    filename_suffix = "msh" if mesh_file_type[:3] == "msh" else mesh_file_type

    preserve_msh = msh_filename is not None
    if msh_filename is None:
        with tempfile.NamedTemporaryFile(suffix="." + filename_suffix) as handle:
            msh_filename = handle.name

    gmsh_executable = gmsh_path if gmsh_path is not None else _get_gmsh_exe()

    args = [
        f"-{dim}",
        geo_filename,
        "-format",
        mesh_file_type,
        "-bin",
        "-o",
        msh_filename,
    ] + extra_gmsh_arguments

    # https://stackoverflow.com/a/803421/353337
    try:
        p = subprocess.Popen(
            [gmsh_executable] + args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        )
    except FileNotFoundError:
        print("Is gmsh installed?")
        raise

    if verbose:
        while True:
            line = p.stdout.readline()
            if not line:
                break
            print(line.decode("utf-8"), end="")

    p.communicate()
    assert p.returncode == 0, "Gmsh exited with error (return code {}).".format(
        p.returncode
    )

    mesh = meshio.read(msh_filename)

    if remove_faces:
        # Only keep the cells of highest topological dimension; discard faces and such.
        two_d_cells = {"triangle", "quad"}
        three_d_cells = {
            "tetra",
            "hexahedron",
            "wedge",
            "pyramid",
            "penta_prism",
            "hexa_prism",
        }
        if any(k in mesh.cells for k in three_d_cells):
            keep_keys = three_d_cells.intersection(mesh.cells.keys())
        elif any(k in mesh.cells for k in two_d_cells):
            keep_keys = two_d_cells.intersection(mesh.cells.keys())
        else:
            keep_keys = set(cell_type for cell_type, _ in mesh.cells)

        for key, val in mesh.cell_data.items():
            mesh.cell_data[key] = [
                d for d, c in zip(val, mesh.cells) if c[0] in keep_keys
            ]
        mesh.cells = [c for c in mesh.cells if c[0] in keep_keys]

    if prune_vertices:
        # Make sure to include only those vertices which belong to a cell.
        ncells = numpy.concatenate([numpy.concatenate(c) for _, c in mesh.cells])
        uvertices, uidx = numpy.unique(ncells, return_inverse=True)

        k = 0
        cells = []
        for key, cellblock in mesh.cells:
            n = numpy.prod(cellblock.shape)
            cells.append(meshio.Cells(key, uidx[k : k + n].reshape(cellblock.shape)))
            k += n
        mesh.cells = cells

        mesh.points = mesh.points[uvertices]
        for key in mesh.point_data:
            mesh.point_data[key] = mesh.point_data[key][uvertices]

    # clean up
    if preserve_msh:
        print(f"\nmsh file: {msh_filename}")
    else:
        os.remove(msh_filename)
    if preserve_geo:
        print(f"\ngeo file: {geo_filename}")
    else:
        os.remove(geo_filename)

    if (
        prune_z_0
        and mesh.points.shape[1] == 3
        and numpy.all(numpy.abs(mesh.points[:, 2]) < 1.0e-13)
    ):
        mesh.points = mesh.points[:, :2]

    return mesh



import pygmsh


def test(lcar=1.0):
    geom = pygmsh.built_in.Geometry()
    poly = geom.add_polygon(
        [[0.0, 0.0, 0.0], [1.0, 0.0, 0.0], [1.0, 1.0, 0.0], [0.0, 1.0, 0.0]], lcar
    )

    geom.set_transfinite_surface(poly.surface, size=[11, 9])

    mesh = pygmsh.generate_mesh(geom, geo_filename="transfinite.geo")
    assert len(mesh.cells_dict["triangle"]) == 10 * 8 * 2
    return mesh


if __name__ == "__main__":
    import meshio

    meshio.write("transfinite.vtu", test())



import numpy as np

import pygmsh
from helpers import compute_volume


def test(radius=1.0):
    geom = pygmsh.built_in.Geometry()

    R = [
        pygmsh.rotation_matrix(np.eye(1, 3, d)[0], theta)
        for d, theta in enumerate(np.pi / np.array([2.0, 3.0, 5]))
    ]

    geom.add_circle([7.0, 11.0, 13.0], radius, 0.1, R[0] @ R[1] @ R[2])

    ref = np.pi * radius ** 2
    mesh = pygmsh.generate_mesh(geom)
    assert np.isclose(compute_volume(mesh), ref, rtol=1e-2)
    return mesh


if __name__ == "__main__":
    import meshio

    meshio.write("circle_transformed.vtk", *test())



import numpy as np

import pygmsh
from helpers import compute_volume


def test():
    geom = pygmsh.built_in.Geometry()

    # internal radius of torus
    irad = 0.15
    # external radius of torus
    orad = 0.27

    Z_pos = (irad + orad) * np.concatenate(
        [+np.ones(8), -np.ones(8), +np.ones(8), -np.ones(8)]
    )

    Alpha = np.concatenate(
        [
            np.arange(8) * np.pi / 4.0,
            np.arange(8) * np.pi / 4.0 + np.pi / 16.0,
            np.arange(8) * np.pi / 4.0,
            np.arange(8) * np.pi / 4.0 + np.pi / 16.0,
        ]
    )

    A1 = (
        (irad + orad)
        / np.tan(np.pi / 8.0)
        * np.concatenate(
            [1.6 * np.ones(8), 1.6 * np.ones(8), 1.9 * np.ones(8), 1.9 * np.ones(8)]
        )
    )

    for alpha, a1, z in zip(Alpha, A1, Z_pos):
        # Rotate torus to the y-z-plane.
        R1 = pygmsh.rotation_matrix([0.0, 1.0, 0.0], 0.5 * np.pi)
        R2 = pygmsh.rotation_matrix([0.0, 0.0, 1.0], alpha)
        x0 = np.array([a1, 0.0, 0.0])
        x1 = np.array([0.0, 0.0, z])
        # First rotate to y-z-plane, then move out to a1, rotate by angle
        # alpha, move up by z.
        #
        #    xnew = R2*(R1*x+x0) + x1
        #
        geom.add_torus(
            irad=irad, orad=orad, lcar=0.1, R=np.dot(R2, R1), x0=np.dot(R2, x0) + x1
        )

    geom.add_box(-1.0, 1.0, -1.0, 1.0, -1.0, 1.0, lcar=0.3)

    ref = len(A1) * 2 * np.pi ** 2 * orad * irad ** 2 + 2.0 ** 3
    mesh = pygmsh.generate_mesh(geom)
    assert np.isclose(compute_volume(mesh), ref, rtol=2e-2)
    return mesh


if __name__ == "__main__":
    import meshio

    meshio.write("torus_crowd.vtu", test())



from .volume_base import VolumeBase


class Box(VolumeBase):
    """
    Creates a box.

    Parameters
    ----------
    x0 : array-like[3]
        List containing the x, y, z values of the start point.
    extends : array-like[3]
        List of the 3 extents of the box edges.
    char_length : float
        Characteristic length of the mesh elements of this polygon.
    """

    def __init__(self, x0, extents, char_length=None):
        super().__init__()

        assert len(x0) == 3
        assert len(extents) == 3

        self.x0 = x0
        self.extents = extents
        self.char_length = char_length

        args = list(x0) + list(extents)
        args = ", ".join([f"{arg}" for arg in args])

        self.code = "\n".join(
            [f"{self.id} = newv;", f"Box({self.id}) = {{{args}}};"]
            + self.char_length_code(char_length)
        )
        return



from .line_loop import LineLoop
from .surface_base import SurfaceBase


class PlaneSurface(SurfaceBase):
    """
    Creates a plane surface.

    Parameters
    ----------
    line_loop : Object
        Each unique line in the line loop will be used
        for the surface construction.
    holes : list
        List of line loops that represents polygon holes.

    Notes
    -----
    The first line loop defines the exterior boundary of the surface;
    all other line loops define holes in the surface.

    A line loop defining a hole should not have any lines in
    common with the exterior line loop (in which case it is not
    a hole, and the two surfaces should be defined separately).

    Likewise, a line loop defining a hole should not have any lines
    in common with another line loop defining a hole in the same
    surface (in which case the two line loops should be combined).
    """

    def __init__(self, line_loop, holes=None):
        super().__init__()

        assert isinstance(line_loop, LineLoop)
        self.line_loop = line_loop

        if holes is None:
            holes = []

        # The input holes are either line loops or entities that contain line
        # loops (like polygons).
        self.holes = [h if isinstance(h, LineLoop) else h.line_loop for h in holes]

        line_loops = [self.line_loop] + self.holes
        self.code = "\n".join(
            [
                f"{self.id} = news;",
                "Plane Surface({}) = {{{}}};".format(
                    self.id, ",".join([ll.id for ll in line_loops])
                ),
            ]
        )
        self.num_edges = len(self.line_loop) + sum(len(h) for h in self.holes)
        return



from .surface_base import SurfaceBase


class Disk(SurfaceBase):
    """
    Creates a disk.

    Parameters
    ----------
    x0 : array-like[3]
        The 3 coordinates of the center of the disk face.
    radius0 : float
        Radius value of the disk.
    radius1 : float
        Radius along Y, leading to an ellipse.
    char_length : float
        Characteristic length of the mesh elements of this polygon.
    """

    def __init__(self, x0, radius0, radius1=None, char_length=None):
        super().__init__()

        assert len(x0) == 3
        if radius1 is not None:
            assert radius0 >= radius1

        self.x0 = x0
        self.radius0 = radius0
        self.radius1 = radius1
        self.char_length = char_length

        args = list(x0) + [radius0]
        if radius1 is not None:
            args.append(radius1)

        args = ", ".join([f"{arg}" for arg in args])

        self.code = "\n".join(
            [f"{self.id} = news;", f"Disk({self.id}) = {{{args}}};"]
            + self.char_length_code(char_length)
        )
        return



import pygmsh
from helpers import compute_volume


def test():
    geom = pygmsh.built_in.Geometry()

    poly = geom.add_polygon(
        [[0.0, 0.5, 0.0], [1.0, 0.5, 0.0], [1.0, 1.0, 0.0], [0.0, 1.0, 0.0]], lcar=0.05
    )

    geom.symmetry(poly, [0.0, 1.0, 0.0, -0.5])

    mesh = pygmsh.generate_mesh(geom)
    ref = 1.0
    assert abs(compute_volume(mesh) - ref) < 1.0e-2 * ref
    return mesh


if __name__ == "__main__":
    import meshio

    meshio.write("symmetry.vtk", test())



"""Creates regular cube mesh by extrusion.
"""
import pygmsh
from helpers import compute_volume


def test():
    x = 5
    y = 4
    z = 3
    x_layers = 10
    y_layers = 5
    z_layers = 3
    geom = pygmsh.opencascade.Geometry()
    p = geom.add_point([0, 0, 0], 1)
    _, l, _ = geom.extrude(p, [x, 0, 0], num_layers=x_layers)
    _, s, _ = geom.extrude(l, [0, y, 0], num_layers=y_layers)
    geom.extrude(s, [0, 0, z], num_layers=z_layers)
    mesh = pygmsh.generate_mesh(geom)

    ref_vol = x * y * z
    assert abs(compute_volume(mesh) - ref_vol) < 1.0e-2 * ref_vol

    # Each grid-cell from layered extrusion will result in 6 tetrahedrons.
    ref_tetras = 6 * x_layers * y_layers * z_layers
    assert len(mesh.cells_dict["tetra"]) == ref_tetras

    return mesh


if __name__ == "__main__":
    import meshio

    meshio.write("cube.vtu", test())



from .surface_base import SurfaceBase


class CompoundSurface(SurfaceBase):
    """
    Generates the Compound Surface Gmsh function.
    Creates a compound surface from several elementary surfaces.
    When meshed, a compound surface will be reparametrized as
    a single surface, whose mesh can thus cross internal boundaries.
    Compound surfaces are mostly useful for remeshing discrete models.

    Parameters
    ----------
    surfaces : array-like[N]
        Contains the identification number of the elementary surfaces
        that should be reparametrized as a single surface.
    """

    def __init__(self, surfaces):
        super().__init__()
        self.num_edges = sum(s.num_edges for s in surfaces)

        self.surfaces = surfaces

        self.code = "\n".join(
            [
                f"{self.id} = news;",
                "Compound Surface({}) = {{{}}};".format(
                    self.id, ",".join([s.id for s in surfaces])
                ),
            ]
        )
        return



from .ball import Ball


class Ellipsoid(Ball):
    """
    Creates an ellipsoid.

    Parameters
    ----------
    center: array-like[3]
        Center of the ellipsoid.
    radii: array-like[3]
        The radii of the ellipsoid along the semi-axes
    x0: float
        If specified and `x0 > -1`, the ellipsoid is cut off at `x0*radius`
        parallel to the y-z plane.
    x1: float
        If specified and `x1 < +1`, the ellipsoid is cut off at `x1*radius`
        parallel to the y-z plane.
    alpha: float
        If specified and `alpha < 2*pi`, the points between `alpha` and
        `2*pi` w.r.t. to the x-y plane are not part of the object.
    char_length: float
        If specified, sets the `Characteristic Length` property.
    """

    def __init__(self, center, radii, char_length=None, **kwargs):
        self.center = center
        self.radii = radii
        self.char_length = char_length

        # Instantiate the ball and later stretch it
        Ball.__init__(self, center, 1.0, char_length=None, **kwargs)

        def as_str(point):
            return ", ".join(str(x) for x in point)

        self.code = "\n".join(
            [
                self.code,
                "Dilate {{{{{}}}, {{{}}}}} {{ Volume{{{}}}; }}".format(
                    as_str(center), as_str(radii), self.id
                ),
            ]
            + self.char_length_code(char_length)
        )
        return



from .line_loop import LineLoop


class Surface:
    """
    Generates a Surface or Rules Surfaces.

    Parameters
    ----------
    line_loop : Object
        LineLoop object that contains all the Line objects for the
        loop construction.
    api_level : integer
        If larger than 2 a Surface will be constructed, otherwise
        a Ruled Surface will be constructed instead.

    Notes
    -----
    With the built-in kernel, the first line loop should be composed of
    either three or four elementary lines.

    With the built-in kernel, the optional In Sphere argument forces the
    surface to be a spherical patch (the extra parameter gives the
    identification number of the center of the sphere).
    """

    _ID = 0
    num_edges = 0
    dimension = 2

    def __init__(self, line_loop, api_level=2):
        assert isinstance(line_loop, LineLoop)

        self.line_loop = line_loop

        self.id = f"rs{Surface._ID}"
        Surface._ID += 1

        # `Ruled Surface` was deprecated in Gmsh 3 in favor of `Surface`.
        name = "Surface" if api_level > 2 else "Ruled Surface"

        self.code = "\n".join(
            [f"{self.id} = news;", f"{name}({self.id}) = {{{self.line_loop.id}}};"]
        )
        self.num_edges = len(line_loop)
        return



from .line_base import LineBase


class CompoundLine(LineBase):
    """
    Creates a compound line from several elementary lines.
    When meshed, a compound line will be reparametrized as
    a single line, whose mesh can thus cross internal boundaries.

    Parameters
    ----------
    lines : array-like[N]
        Contains the identification number of the elementary lines
        that should be reparametrized as a single line.
    """

    def __init__(self, lines):
        super().__init__()

        self.lines = lines

        self.code = "\n".join(
            [
                f"{self.id} = newl;",
                "Compound Line({}) = {{{}}};".format(
                    self.id, ",".join([l.id for l in self.lines])
                ),
            ]
        )
        return



from .volume_base import VolumeBase


class Torus(VolumeBase):
    """
    Creates a torus.

    center : array-like[3]
        The 3 coordinates of its center.
    radius0 : float
        Inner radius.
    radius1 : float
        Outer radius.
    alpha : float
        Defines the angular opening.
    char_length : float
        Characteristic length of the mesh elements of this polygon.
    """

    def __init__(self, center, radius0, radius1, alpha=None, char_length=None):
        super().__init__()

        assert len(center) == 3

        self.center = center
        self.radius0 = radius0
        self.radius1 = radius1
        self.alpha = alpha
        self.char_length = char_length

        args = list(center) + [radius0] + [radius1]
        if alpha is not None:
            args.append(alpha)
        args = ", ".join([f"{arg}" for arg in args])

        self.code = "\n".join(
            [f"{self.id} = newv;", f"Torus({self.id}) = {{{args}}};"]
            + self.char_length_code(char_length)
        )
        return



import numpy as np

import pygmsh
from helpers import compute_volume


def test():
    geom = pygmsh.built_in.Geometry()

    X0 = np.array(
        [[+0.0, +0.0, 0.0], [+0.5, +0.3, 0.1], [-0.5, +0.3, 0.1], [+0.5, -0.3, 0.1]]
    )

    R = np.array([0.1, 0.2, 0.1, 0.14])

    holes = [
        geom.add_ball(x0, r, with_volume=False, lcar=0.2 * r).surface_loop
        for x0, r in zip(X0, R)
    ]

    # geom.add_box(
    #         -1, 1,
    #         -1, 1,
    #         -1, 1,
    #         lcar=0.2,
    #         holes=holes
    #         )

    geom.add_ball([0, 0, 0], 1.0, lcar=0.2, holes=holes)

    # geom.add_physical_volume(ball, label="cheese")

    ref = 4.07064892966291
    mesh = pygmsh.generate_mesh(geom)
    assert abs(compute_volume(mesh) - ref) < 2.0e-2 * ref
    return mesh


if __name__ == "__main__":
    import meshio

    meshio.write("swiss_cheese.vtu", test())



import numpy as np

import pygmsh


def test():
    # Characteristic length
    lcar = 1e-1

    # Coordinates of lower-left and upper-right vertices of a square domain
    xmin = 0.0
    xmax = 5.0
    ymin = 0.0
    ymax = 5.0

    # Vertices of a square hole
    squareHoleCoordinates = np.array([[1, 1, 0], [4, 1, 0], [4, 4, 0], [1, 4, 0]])

    # Create geometric object
    geom = pygmsh.built_in.Geometry()

    # Create square hole
    squareHole = geom.add_polygon(squareHoleCoordinates, lcar, make_surface=False)

    # Create square domain with square hole
    geom.add_rectangle(xmin, xmax, ymin, ymax, 0.0, lcar, holes=[squareHole.line_loop])

    mesh = pygmsh.generate_mesh(geom, extra_gmsh_arguments=["-order", "2"])
    # TODO support for volumes of triangle6
    # ref = 16.0
    # from helpers import compute_volume
    # assert abs(compute_volume(points, cells) - ref) < 1.0e-2 * ref
    return mesh


if __name__ == "__main__":
    import meshio

    meshio.write("hole_in_square.vtu", test())



import pygmsh


def test(lcar=0.5):
    geom = pygmsh.built_in.Geometry()
    poly = geom.add_polygon(
        [[0.0, 0.0, 0.0], [1.0, 0.0, 0.0], [1.0, 1.0, 0.0], [0.0, 1.0, 0.0]], lcar
    )

    top, volume, lat = geom.extrude(poly.surface, [0, 0, 2])

    geom.add_physical(poly.surface, label="bottom")
    geom.add_physical(top, label="top")
    geom.add_physical(volume, label="volume")
    geom.add_physical(lat, label="lat")
    geom.add_physical(poly.lines[0], label="line")

    mesh = pygmsh.generate_mesh(geom, geo_filename="physical.geo")
    return mesh


if __name__ == "__main__":
    import meshio

    meshio.write("physical.vtu", test())



import pytest

import pygmsh
from helpers import compute_volume


@pytest.mark.skipif(pygmsh.get_gmsh_major_version() < 3, reason="requires Gmsh >= 3")
def test():
    geom = pygmsh.opencascade.Geometry()

    rectangle = geom.add_rectangle(
        [-1.0, -1.0, 0.0], 2.0, 2.0, corner_radius=0.2, char_length=0.05
    )
    disk1 = geom.add_disk([-1.2, 0.0, 0.0], 0.5)
    disk2 = geom.add_disk([+1.2, 0.0, 0.0], 0.5, 0.3)
    union = geom.boolean_union([rectangle, disk1, disk2])

    disk3 = geom.add_disk([0.0, -0.9, 0.0], 0.5)
    disk4 = geom.add_disk([0.0, +0.9, 0.0], 0.5)
    flat = geom.boolean_difference([union], [disk3, disk4])

    geom.extrude(flat, [0, 0, 0.3])

    ref = 1.1742114942
    mesh = pygmsh.generate_mesh(geom)
    assert abs(compute_volume(mesh) - ref) < 1.0e-2 * ref
    return mesh


if __name__ == "__main__":
    import meshio

    meshio.write("opencascade_extrude.vtu", test())



from .point_base import PointBase


class Point(PointBase):
    """
    Creates an elementary point.

    x : array-like[3]
        Give the three X, Y and Z coordinates of the
        point in the three-dimensional Euclidean space.
    lcar : float
        The prescribed mesh element size at this point.
    """

    def __init__(self, x, lcar=None):
        super().__init__()

        self.x = x
        self.lcar = lcar

        # Points are always 3D in gmsh
        args = (x[0], x[1], x[2]) if lcar is None else (x[0], x[1], x[2], lcar)
        fmt = ", ".join(len(args) * ["{!r}"])

        self.code = "\n".join(
            [
                f"{self.id} = newp;",
                ("Point({}) = {{" + fmt + "}};").format(self.id, *args),
            ]
        )
        return



from .line_base import LineBase
from .point import Point


class Spline(LineBase):
    """
    With the built-in geometry kernel this constructs a Catmull-Rom spline.

    Parameters
    ----------
    points : list
        List containing Point objects
    """

    def __init__(self, points):
        super().__init__()

        for c in points:
            assert isinstance(c, Point)
        assert len(points) > 1

        self.points = points

        self.code = "\n".join(
            [
                f"{self.id} = newl;",
                "Spline({}) = {{{}}};".format(
                    self.id, ", ".join([c.id for c in self.points])
                ),
            ]
        )
        return



class VolumeBase:
    """
    Increments the Line ID every time a new object
    is created that inherits from LineBase.

    Parameters
    ----------
    id0 : str
        If no unique ID is given, the object global is incremented.
    """

    _ID = 0
    dimension = 3

    def __init__(self, id0=None):
        if id0:
            self.id = id0
        else:
            self.id = f"vol{VolumeBase._ID}"
            VolumeBase._ID += 1
        return



import numpy as np

import pygmsh


def test():
    geom = pygmsh.built_in.Geometry()
    p0 = geom.add_point((0.0, 0.0, 0.0), lcar=1.0)
    p1 = geom.add_point((2.0, 0.0, 0.0), lcar=1.0)
    p2 = geom.add_point((0.0, 1.0, 0.0), lcar=1.0)
    p3 = geom.add_point((2.0, 1.0, 0.0), lcar=1.0)
    l0 = geom.add_line(p0, p1)
    l1 = geom.add_line(p1, p3)
    l2 = geom.add_line(p3, p2)
    l3 = geom.add_line(p2, p0)
    ll0 = geom.add_line_loop((l0, l1, l2, l3))
    rs0 = geom.add_surface(ll0)
    geom.set_transfinite_lines([l3, l1], 3, progression=1)
    geom.set_transfinite_lines([l2, l0], 3, progression=1)
    geom.set_transfinite_surface(rs0)
    geom.set_recombined_surfaces([rs0])

    mesh = pygmsh.generate_mesh(geom)

    assert "quad" in mesh.cells_dict.keys()
    ref = np.array([[0, 4, 8, 7], [7, 8, 6, 2], [4, 1, 5, 8], [8, 5, 3, 6]])
    assert np.array_equal(ref, mesh.cells_dict["quad"])

    return mesh


if __name__ == "__main__":
    import meshio

    meshio.write("rectangle_structured.vtk", test())



import os

from setuptools import find_packages, setup

# https://packaging.python.org/single_source_version/
base_dir = os.path.abspath(os.path.dirname(__file__))
about = {}
with open(os.path.join(base_dir, "pygmsh", "__about__.py"), "rb") as f:
    exec(f.read(), about)


setup(
    name="pygmsh",
    version=about["__version__"],
    author=about["__author__"],
    author_email=about["__author_email__"],
    packages=find_packages(),
    description="Python frontend for Gmsh",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url=about["__website__"],
    project_urls={
        "Code": "https://github.com/nschloe/meshio",
        "Documentation": "https://pygmsh.readthedocs.org/en/latest",
        "Issue tracker": "https://github.com/nschloe/meshio/issues",
    },
    license=about["__license__"],
    platforms="any",
    install_requires=["meshio >=4.0, <5.0", "numpy >= 1.9"],
    python_requires=">=3.6",
    keywords=["mesh", "gmsh", "mesh generation", "mathematics"],
    classifiers=[
        about["__status__"],
        about["__license__"],
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering :: Mathematics",
    ],
)



from .. import built_in


class VolumeBase(built_in.volume_base.VolumeBase):
    """
    Increments the Volume ID every time a new volume object
    is created. Inherits from built_in VolumeBase.
    """

    _ID = 0
    dimension = 3

    def __init__(self, is_list=False, id0=None):
        super().__init__(id0=id0)

        self.is_list = is_list
        if is_list:
            self.id += "[]"
        return

    def char_length_code(self, char_length):
        if char_length is None:
            return []

        return [
            f"pts_{self.id}[] = PointsOf{{Volume{{{self.id}}};}};",
            f"Characteristic Length{{pts_{self.id}[]}} = {char_length};",
        ]



__version__ = "6.0.5"
__author__ = "Nico Schlömer"
__author_email__ = "nico.schloemer@gmail.com"
__copyright__ = f"Copyright (c) 2013-2020, {__author__} <{__author_email__}>"
__website__ = "https://github.com/nschloe/pygmsh"
__license__ = "License :: OSI Approved :: MIT License"
__status__ = "Development Status :: 5 - Production/Stable"



from .volume_base import VolumeBase


class Volume(VolumeBase):
    """
    Creates a volume.

    Parameters
    ----------
    surface_loop : list
        Contain the identification numbers of all the surface
        loops defining the volume.
    holes : list
        List containing surface loop objects that represents polygon holes.

    Notes
    -----
    The first surface loop defines the exterior boundary of the volume;
    all other surface loops define holes in the volume.

    A surface loop defining a hole should not have any surfaces in common
    with the exterior surface loop (in which case it is not a hole,
    and the two volumes should be defined separately).

    Likewise, a surface loop defining a hole should not have any surfaces
    in common with another surface loop defining a hole in the same volume
    (in which case the two surface loops should be combined).
    """

    def __init__(self, surface_loop, holes=None):
        super().__init__()

        if holes is None:
            holes = []

        self.surface_loop = surface_loop
        self.holes = holes

        surface_loops = [surface_loop] + holes

        self.code = "\n".join(
            [
                f"{self.id} = newv;",
                "Volume({}) = {{{}}};".format(
                    self.id, ", ".join([s.id for s in surface_loops])
                ),
            ]
        )
        return
